# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import pathlib

__version__: str = (pathlib.Path(__file__).parent / "VERSION").read_text().strip()
__version_info__: typing.Tuple[int, ...] = tuple(map(int, __version__.split(".")[:3]))
__author__ = "UAVCAN Consortium"
__email__ = "consortium@uavcan.org"
__copyright__ = f"Copyright (c) 2020 {__author__} <{__email__}>"
__license__ = "MIT"

from .main import main as main, subcommand as subcommand, Purser as Purser, pass_purser as pass_purser
from .main import asynchronous as asynchronous
from . import cmd as cmd



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import os
import sys
import typing
import logging
import click
import yakut
from yakut.param.transport import transport_factory_option, TransportFactory, Transport
from yakut.param.formatter import formatter_factory_option, FormatterFactory, Formatter
from yakut.param.node import node_factory_option, NodeFactory


_logger = logging.getLogger(__name__.replace("__", ""))
_LOG_FORMAT = "%(asctime)s %(process)07d %(levelname)-3.3s %(name)s: %(message)s"
logging.basicConfig(format=_LOG_FORMAT)  # Using the default log level; it will be overridden later.


class Purser:
    def __init__(
        self,
        formatter_factory: FormatterFactory,
        transport_factory: TransportFactory,
        node_factory: NodeFactory,
    ) -> None:
        self._f_formatter = formatter_factory
        self._f_transport = transport_factory
        self._f_node = node_factory

        self._transport: typing.Optional[Transport] = None
        self._node: typing.Optional[object] = None

    def make_formatter(self) -> Formatter:
        return self._f_formatter()

    def get_transport(self) -> Transport:
        if self._transport is None:
            self._transport = self._f_transport()
        if self._transport is not None:
            return self._transport
        click.get_current_context().fail("Transport not configured")

    def get_node(self, name_suffix: str, allow_anonymous: bool) -> object:
        if self._node is None:
            tr = self.get_transport()
            self._node = self._f_node(tr, name_suffix=name_suffix, allow_anonymous=allow_anonymous)
        return self._node


pass_purser = click.make_pass_decorator(Purser)


class AbbreviatedGroup(click.Group):
    def get_command(self, ctx: click.Context, cmd_name: str) -> typing.Optional[click.Command]:
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        if len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(f"Abbreviated command {cmd_name!r} is ambiguous. Possible matches: {list(matches)}")

    def resolve_command(
        self, ctx: click.Context, args: typing.List[typing.Any]
    ) -> typing.Tuple[str, click.Command, typing.List[typing.Any]]:
        """
        This is a workaround for this bug in v7: https://github.com/pallets/click/issues/1422.

        If this is not overridden, then abbreviated commands cause the automatic envvar prefix to be constructed
        incorrectly, such that instead of the full command name the abbreviated name is used.
        For example, if the user invokes `yakut comp` meaning `yakut compile`,
        the auto-constructed envvar prefix would be `YAKUT_COMP_` instead of `YAKUT_COMPILE_`.
        """
        _, cmd, out_args = super().resolve_command(ctx, args)
        return cmd.name, cmd, out_args


_ENV_VAR_PATH = "YAKUT_PATH"


@click.command(
    cls=AbbreviatedGroup,
    context_settings={
        "max_content_width": click.get_terminal_size()[0],
    },
)
@click.version_option(version=yakut.__version__)
@click.option("--verbose", "-v", count=True, help="Show verbose log messages. Specify twice for extra verbosity.")
@click.option(
    "--path",
    "-P",
    multiple=True,
    type=click.Path(resolve_path=True),
    help=f"""
In order to use compiled DSDL namespaces,
the directories that contain compilation outputs need to be specified using this option.
The current working directory does not need to be specified explicitly.

Examples:

\b
    yakut  --path ../public_regulated_data_types  --path ~/my_namespaces  pub ...

\b
    export {_ENV_VAR_PATH}="../public_regulated_data_types:~/my_namespaces"
    yakut pub ...
""",
)
@formatter_factory_option
@transport_factory_option
@node_factory_option
@click.pass_context
def main(
    ctx: click.Context,
    verbose: int,
    path: typing.Tuple[str, ...],
    formatter_factory: FormatterFactory,
    transport_factory: TransportFactory,
    node_factory: NodeFactory,
) -> None:
    """
    \b
         __   __   _______   __   __   _______   _______   __   __
        |  | |  | /   _   ` |  | |  | /   ____| /   _   ` |  ` |  |
        |  | |  | |  |_|  | |  | |  | |  |      |  |_|  | |   `|  |
        |  |_|  | |   _   | `  `_/  / |  |____  |   _   | |  |`   |
        `_______/ |__| |__|  `_____/  `_______| |__| |__| |__| `__|
            |      |            |         |      |         |
        ----o------o------------o---------o------o---------o-------

    Yakut is a cross-platform command-line utility for diagnostics and management of UAVCAN networks.
    It is designed for use either directly by humans or from automation scripts.
    Ask questions at https://forum.uavcan.org

    Any long option can be provided via environment variable prefixed with `YAKUT_`
    such that an option `--foo-bar` for command `baz`, if not provided as a command-line argument,
    will be read from `YAKUT_BAZ_FOO_BAR`.

    Any command can be abbreviated arbitrarily as long as the resulting abridged name is not ambiguous.
    For example, `publish`, `publ` and `pub` are all valid and equivalent.
    """
    _configure_logging(verbose)  # This should be done in the first order to ensure that we log things correctly.

    for p in (os.getcwd(), *path):
        _logger.debug("New path item: %s", p)
        sys.path.append(str(p))

    ctx.obj = Purser(
        formatter_factory=formatter_factory,
        transport_factory=transport_factory,
        node_factory=node_factory,
    )


subcommand: typing.Callable[..., typing.Callable[..., typing.Any]] = main.command  # type: ignore


def asynchronous(f: typing.Callable[..., typing.Awaitable[typing.Any]]) -> typing.Callable[..., typing.Any]:
    import asyncio
    from functools import update_wrapper

    def proxy(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return update_wrapper(proxy, f)


def _configure_logging(verbosity_level: int) -> None:
    log_level = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
    }.get(verbosity_level or 0, logging.DEBUG)

    logging.root.setLevel(log_level)

    try:
        import coloredlogs  # type: ignore

        # The level spec applies to the handler, not the root logger! This is different from basicConfig().
        coloredlogs.install(level=log_level, fmt=_LOG_FORMAT)
    except Exception as ex:  # pylint: disable=broad-except
        _logger.exception("Could not set up coloredlogs: %r", ex)  # pragma: no cover



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import http
import shutil
import typing
import logging
import zipfile
import tempfile
from pathlib import Path
import pyuavcan
import click
import yakut
from yakut.paths import DEFAULT_PUBLIC_REGULATED_DATA_TYPES_ARCHIVE_URI


_NAME = "compile"

_logger = logging.getLogger(__name__)


def make_usage_suggestion(root_namespace_name: typing.Optional[str]) -> str:
    """
    When a command is unable to find a compiled DSDL package, this helper can be used to construct a
    human-friendly suggestion on how to resolve the problem.
    """
    root_namespace_name = root_namespace_name or "<namespace>"
    root_namespace_name = root_namespace_name.split(".")[0]  # Transform: `uavcan.node` --> `uavcan`
    return f"Run `yakut {_NAME} <path>/{root_namespace_name}` to compile DSDL namespace {root_namespace_name!r}"


@yakut.subcommand(
    name=_NAME,
    help=f"""
Compile DSDL namespaces for use by Yakut.
This needs to be done before using any data types with pub/sub/call and other commands.

The command accepts a list of sources where each element is either a local path
or an URI pointing to the source DSDL root namespace(s).

If a source is a local path, it must point to a local DSDL root namespace directory or to a local archive containing
DSDL root namespace directories at the top level.
If the value is an URI, it must point to an archive containing DSDL root namespace directories at the top level
(this is convenient for generating packages from namespaces hosted in public repositories, e.g., on GitHub).

See also: top-level option `--path` and related environment variable `YAKUT_PATH`.

Example path: ~/uavcan/public_regulated_data_types/uavcan/

Example URI: {DEFAULT_PUBLIC_REGULATED_DATA_TYPES_ARCHIVE_URI}

Example command that compiles the root namespace `~/namespace` which depends on the public regulated types:

\b
    yakut {_NAME}  ~/namespace  --lookup {DEFAULT_PUBLIC_REGULATED_DATA_TYPES_ARCHIVE_URI}
""",
)
@click.argument("source", nargs=-1, required=True, type=str)
@click.option(
    "--lookup",
    "-L",
    multiple=True,
    type=str,
    metavar="SOURCE",
    help=f"""
This is like the sources except that the specified DSDL root namespace(s) will be used only for looking up dependencies.
Both local directories and URIs are accepted.
If a DSDL root namespace is specified as an input, it is automatically added to the look-up list as well.
""",
)
@click.option(
    "--output",
    "-O",
    type=click.Path(file_okay=False, dir_okay=True, writable=True, resolve_path=True),
    help=f"""
Path to the directory where the compilation outputs will be stored.
If not specified, defaults to the current working directory.
Existing packages will be overwritten entirely.
""",
)
@click.option(
    "--allow-unregulated-fixed-port-id",
    is_flag=True,
    help="""
Instruct the DSDL front-end to accept unregulated data types with fixed port identifiers.
Make sure you understand the implications before using this option.
If not sure, ask for advice at https://forum.uavcan.org.
""",
)
def compile_(
    source: typing.Tuple[str, ...],
    lookup: typing.Tuple[str, ...],
    output: typing.Union[str, Path, None],
    allow_unregulated_fixed_port_id: bool,
) -> None:
    output = Path(output or Path.cwd()).resolve()
    _logger.info("Destination: %r", str(output))

    src_dirs: typing.List[Path] = []
    for location in source:
        src_dirs += _fetch_root_namespace_dirs(location)
    _logger.info("Source namespace dirs: %r", list(map(str, src_dirs)))

    lookup_dirs: typing.List[Path] = []
    for location in lookup:
        lookup_dirs += _fetch_root_namespace_dirs(location)
    _logger.info("Lookup namespace dirs: %r", list(map(str, lookup_dirs)))

    gpi_list = _generate_dsdl_packages(
        source_root_namespace_dirs=src_dirs,
        lookup_root_namespace_dirs=lookup_dirs,
        generated_packages_dir=output,
        allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id,
    )
    for gpi in gpi_list:
        _logger.info("Generated package %r with %d data types at %r", gpi.name, len(gpi.models), str(gpi.path))


def _fetch_root_namespace_dirs(location: str) -> typing.List[Path]:
    if "://" in location:
        dirs = _fetch_archive_dirs(location)
        _logger.info(
            "Resource %r contains the following root namespace directories: %r", location, list(map(str, dirs))
        )
        return dirs
    return [Path(location)]


def _fetch_archive_dirs(archive_uri: str) -> typing.List[Path]:
    """
    Downloads an archive from the specified URI, unpacks it into a temporary directory, and returns the list of
    directories in the root of the unpacked archive.
    """
    import requests  # Takes over 100 ms to import! Having it in the file scope is a performance disaster.

    # TODO: autodetect the type of the archive
    arch_dir = tempfile.mkdtemp(prefix="pyuavcan-cli-dsdl")
    arch_file = str(Path(arch_dir) / "dsdl.zip")

    _logger.info("Downloading the archive from %r into %r...", archive_uri, arch_file)
    response = requests.get(archive_uri)
    if response.status_code != http.HTTPStatus.OK:
        raise RuntimeError(f"Could not download the archive; HTTP error {response.status_code}")
    with open(arch_file, "wb") as f:
        f.write(response.content)

    _logger.info("Extracting the archive into %r...", arch_dir)
    with zipfile.ZipFile(arch_file) as zf:
        zf.extractall(arch_dir)

    (inner,) = [d for d in Path(arch_dir).iterdir() if d.is_dir()]  # Strip the outer layer, we don't need it

    assert isinstance(inner, Path)
    return [d for d in inner.iterdir() if d.is_dir()]


def _generate_dsdl_packages(
    source_root_namespace_dirs: typing.Iterable[Path],
    lookup_root_namespace_dirs: typing.Iterable[Path],
    generated_packages_dir: Path,
    allow_unregulated_fixed_port_id: bool,
) -> typing.Sequence[pyuavcan.dsdl.GeneratedPackageInfo]:
    lookup_root_namespace_dirs = frozenset(list(lookup_root_namespace_dirs) + list(source_root_namespace_dirs))
    generated_packages_dir.mkdir(parents=True, exist_ok=True)

    out: typing.List[pyuavcan.dsdl.GeneratedPackageInfo] = []
    for ns in source_root_namespace_dirs:
        if ns.name.startswith("."):
            _logger.debug("Skipping hidden directory %r", ns)
            continue
        dest_dir = generated_packages_dir / ns.name
        _logger.info(
            "Generating DSDL package %r from root namespace %r with lookup dirs: %r",
            str(dest_dir),
            str(ns),
            list(map(str, lookup_root_namespace_dirs)),
        )
        shutil.rmtree(dest_dir, ignore_errors=True)
        gpi = pyuavcan.dsdl.generate_package(
            root_namespace_directory=ns,
            lookup_directories=list(lookup_root_namespace_dirs),
            output_directory=generated_packages_dir,
            allow_unregulated_fixed_port_id=allow_unregulated_fixed_port_id,
        )
        if gpi is not None:
            out.append(gpi)
    return out



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import enum
import typing
import click


class EnumParam(click.Choice):
    """
    A parameter that allows the user to select one of the enum options.
    The selection is case-insensitive and abbreviations are supported out of the box:
    F, foo, and FOO_BAR are all considered equivalent as long as there are no ambiguities.
    """

    def __init__(self, e: enum.EnumMeta) -> None:
        self._enum = e
        super().__init__(list(e.__members__), case_sensitive=False)

    def convert(
        self,
        value: typing.Union[str, enum.Enum],
        param: typing.Optional[click.Parameter],
        ctx: typing.Optional[click.Context],
    ) -> typing.Any:
        if isinstance(value, enum.Enum):  # This is to support default enum options.
            value = value.name
        assert isinstance(value, str)
        candidates: typing.List[enum.Enum] = [  # type: ignore
            x for x in self._enum if x.name.upper().startswith(value.upper())
        ]
        if len(candidates) == 0:
            raise click.BadParameter(f"Value {value!r} is not a valid choice for {list(self._enum.__members__)}")
        if len(candidates) > 1:
            raise click.BadParameter(f"Value {value!r} is ambiguous; possible matches: {[x.name for x in candidates]}")
        return candidates[0]



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import re
import typing
import logging
import decimal
import importlib
import pyuavcan.dsdl


_NAME_COMPONENT_SEPARATOR = "."


_logger = logging.getLogger(__name__)


def construct_port_id_and_type(spec: str) -> typing.Tuple[int, typing.Type[pyuavcan.dsdl.CompositeObject]]:
    r"""
    Parses a data specifier string of the form ``[port_id.]full_data_type_name.major_version.minor_version``.
    Name separators may be replaced with ``/`` or ``\`` for compatibility with file system paths;
    the version number separators may also be underscores for convenience.
    Raises ValueError, possibly with suggestions, if such type is non-reachable.
    """
    from yakut.cmd.compile import make_usage_suggestion

    port_id, full_name, major, minor = _parse_data_spec(spec)
    name_components = full_name.split(_NAME_COMPONENT_SEPARATOR)
    namespace_components, short_name = name_components[:-1], name_components[-1]
    _logger.debug(
        "Parsed data spec %r: port_id=%r, namespace_components=%r, short_name=%r, major=%r, minor=%r",
        spec,
        port_id,
        namespace_components,
        short_name,
        major,
        minor,
    )

    # Import the generated data type.
    try:
        mod = None
        for comp in namespace_components:
            name = (mod.__name__ + "." + comp) if mod else comp  # type: ignore
            try:
                mod = importlib.import_module(name)
            except ImportError:  # We seem to have hit a reserved word; try with an underscore.
                mod = importlib.import_module(name + "_")
    except ImportError:
        raise ValueError(
            f"The data spec string specifies a non-existent namespace: {spec!r}. "
            f"{make_usage_suggestion(namespace_components[0])}"
        ) from None

    try:
        dtype = getattr(mod, f"{short_name}_{major}_{minor}")
    except AttributeError:
        raise ValueError(f"The data spec string specifies a non-existent short type name: {spec!r}") from None

    if issubclass(dtype, pyuavcan.dsdl.CompositeObject):
        model = pyuavcan.dsdl.get_model(dtype)
        port_id = port_id if port_id is not None else model.fixed_port_id
        if port_id is None:
            raise ValueError(
                f"The data spec does not specify a port ID, "
                f"and a fixed port ID is not defined for the specified data type: {spec!r}"
            )
        return port_id, dtype
    raise ValueError(f"The data spec does not specify a valid type: {spec!r}")


def convert_transfer_metadata_to_builtin(
    transfer: pyuavcan.transport.TransferFrom, **extra_fields: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Any]:
    out = {
        "timestamp": {
            "system": transfer.timestamp.system.quantize(_MICRO),
            "monotonic": transfer.timestamp.monotonic.quantize(_MICRO),
        },
        "priority": transfer.priority.name.lower(),
        "transfer_id": transfer.transfer_id,
        "source_node_id": transfer.source_node_id,
    }
    out.update(extra_fields)
    return {"_metadata_": out}


_MICRO = decimal.Decimal("0.000001")

_RE_SPLIT = re.compile(r"^(?:(\d+)\.)?((?:[a-zA-Z_][a-zA-Z0-9_]*)(?:\.[a-zA-Z_][a-zA-Z0-9_]*)*)[_.](\d+)[_.](\d+)$")
"""
Splits ``123.ns.Type.123.45`` into ``('123', 'ns.Type', '123', '45')``.
Splits     ``ns.Type.123.45`` into ``(None, 'ns.Type', '123', '45')``.
The version separators (the last two) may be underscores.
"""


def _parse_data_spec(spec: str) -> typing.Tuple[typing.Optional[int], str, int, int]:
    r"""
    Transform the provided data spec into: [port-ID], full name, major version, minor version.
    Component separators may be ``/`` or ``\``. Version number separators (the last two) may also be underscores.
    Raises ValueError if non-compliant.
    """
    spec = spec.strip().replace("/", _NAME_COMPONENT_SEPARATOR).replace("\\", _NAME_COMPONENT_SEPARATOR)
    match = _RE_SPLIT.match(spec)
    if match is None:
        raise ValueError(f"Malformed data spec: {spec!r}")
    frag_port_id, frag_full_name, frag_major, frag_minor = match.groups()
    return (int(frag_port_id) if frag_port_id is not None else None), frag_full_name, int(frag_major), int(frag_minor)


def _unittest_parse_data_spec() -> None:
    import pytest

    assert (123, "ns.Type", 12, 34) == _parse_data_spec(" 123.ns.Type.12.34 ")
    assert (123, "ns.Type", 12, 34) == _parse_data_spec("123.ns.Type_12.34")
    assert (123, "ns.Type", 12, 34) == _parse_data_spec("123.ns/Type.12_34")
    assert (123, "ns.Type", 12, 34) == _parse_data_spec("123.ns.Type_12_34")
    assert (123, "ns.Type", 12, 34) == _parse_data_spec(r"123\ns\Type_12_34 ")

    assert (None, "ns.Type", 12, 34) == _parse_data_spec("ns.Type.12.34 ")
    assert (123, "Type", 12, 34) == _parse_data_spec("123.Type.12.34")
    assert (None, "Type", 12, 34) == _parse_data_spec("Type.12.34")
    assert (123, "ns0.sub.Type0", 0, 1) == _parse_data_spec("123.ns0.sub.Type0.0.1")
    assert (None, "ns0.sub.Type0", 255, 255) == _parse_data_spec(r"ns0/sub\Type0.255.255")

    with pytest.raises(ValueError):
        _parse_data_spec("123.ns.Type.12")
    with pytest.raises(ValueError):
        _parse_data_spec("123.ns.Type.12.43.56")
    with pytest.raises(ValueError):
        _parse_data_spec("ns.Type.12")
    with pytest.raises(ValueError):
        _parse_data_spec("ns.Type.12.43.56")
    with pytest.raises(ValueError):
        _parse_data_spec("123.ns.0Type.12.43")



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import pathlib
import click
import yakut


VERSION_AGNOSTIC_DATA_DIR = pathlib.Path(click.get_app_dir("yakut"))
VERSION_SPECIFIC_DATA_DIR = VERSION_AGNOSTIC_DATA_DIR / ("v" + ".".join(map(str, yakut.__version_info__[:2])))

OUTPUT_TRANSFER_ID_MAP_DIR = VERSION_SPECIFIC_DATA_DIR / "output-transfer-id-maps"

OUTPUT_TRANSFER_ID_MAP_MAX_AGE = 60.0  # [second]
"""This is not a path but a related parameter so it's kept here. Files older that this are not used."""

DEFAULT_PUBLIC_REGULATED_DATA_TYPES_ARCHIVE_URI = (
    "https://github.com/UAVCAN/public_regulated_data_types/archive/master.zip"
)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
The YAML library we use is API-unstable at the time of writing. We can't just use the de-facto standard PyYAML
because it's kinda stuck in the past (no ordered dicts, no support for YAML v1.2). This facade shields the
rest of the code from breaking changes in the YAML library API or from migration to another library.
"""

import io
import typing
import decimal
import ruamel.yaml


class YAMLDumper:
    """
    YAML generation facade.
    """

    def __init__(self, explicit_start: bool = False):
        # We need to use the roundtrip representer to retain ordering of mappings, which is important for usability.
        self._impl = ruamel.yaml.YAML(typ="rt")
        # noinspection PyTypeHints
        self._impl.explicit_start = explicit_start
        self._impl.default_flow_style = False

    def dump(self, data: typing.Any, stream: typing.TextIO) -> None:
        self._impl.dump(data, stream)

    def dumps(self, data: typing.Any) -> str:
        s = io.StringIO()
        self.dump(data, s)
        return s.getvalue()


class YAMLLoader:
    """
    YAML parsing facade.
    Natively represents decimal.Decimal as floats in the output.
    """

    def __init__(self) -> None:
        self._impl = ruamel.yaml.YAML()

    def load(self, text: str) -> typing.Any:
        return self._impl.load(text)


def _represent_decimal(self: ruamel.yaml.BaseRepresenter, data: decimal.Decimal) -> ruamel.yaml.ScalarNode:
    if data.is_finite():
        s = str(_POINT_ZERO_DECIMAL + data)  # The zero addition is to force float-like string representation
    elif data.is_nan():
        s = ".nan"
    elif data.is_infinite():
        s = ".inf" if data > 0 else "-.inf"
    else:
        assert False
    return self.represent_scalar("tag:yaml.org,2002:float", s)


ruamel.yaml.add_representer(decimal.Decimal, _represent_decimal, representer=ruamel.yaml.RoundTripRepresenter)

_POINT_ZERO_DECIMAL = decimal.Decimal("0.0")


def _unittest_yaml() -> None:
    import pytest

    ref = YAMLDumper(explicit_start=True).dumps(
        {
            "abc": decimal.Decimal("-inf"),
            "def": [
                decimal.Decimal("nan"),
                {
                    "qaz": decimal.Decimal("789"),
                },
            ],
        }
    )
    assert (
        ref
        == """---
abc: -.inf
def:
- .nan
- qaz: 789.0
"""
    )
    assert YAMLLoader().load(ref) == {
        "abc": -float("inf"),
        "def": [
            pytest.approx(float("nan"), nan_ok=True),
            {
                "qaz": pytest.approx(789),
            },
        ],
    }
